Module lib_vectors Import lib_finite lib_list_length;

Inductive 
[vect:{n:nat}Type] ElimOver Type
Parameters 
[A:Type]
Theorems
Constructors
[vnil:vect zero]
[vcons:{a:A}{n|nat}{v:vect n}(vect (suc n))];

(* Vectors as an inductive type ... very similar to list *)

Goal {n:nat}{m:vect (suc n)}{P:(vect (suc n))->Type}
	({a:A}{m:vect n}P (vcons a m))->P m;
	intros _;
	Refine vect_elim 
              (nat_elim ([x:nat](vect x)->Type) ([_:vect zero]trueProp)
		([x|nat][_:(vect x)->Type][n = suc x][m:vect n]
		{P:(vect n)->Type}({a:A}{m:vect x}P (vcons a m))->P m));
	Intros; Immed;
	Intros; Refine H;
SaveUnfrozen vect_elim'_lemma;


[vect_rec [C|Type] = vect_elim [n|nat][_:vect n]C];
[vect_iter [C|Type][a:C][f:nat->A->C->C] = 
               vect_rec a [x:A][n|nat][_:vect n]f n x];

[vect_ind [P:{n:nat}(vect n)->Prop] = vect_elim P];


Goal vnil_lemma : {n:nat}{v:vect n}(([n|nat]((nat_elim|([x:nat](vect x)->Prop)
           ([x=zero][v:vect zero](Eq v vnil))
           ([x:nat][_:((vect x)->Prop)][_:(vect (suc x))]trueProp)
           n):((vect n)->Prop))) v);
Refine vect_elim [n|nat]((nat_elim|([x:nat](vect x)->Prop)
           ([x=zero][v:vect zero](Eq v vnil))
           ([x:nat][_:((vect x)->Prop)][_:(vect (suc x))]trueProp)
           n):((vect n)->Prop));
Refine Eq_refl;
Intros;
Immed;
Save vnil_lemma;

Goal vect_zero_eq_vnil : {v:vect zero}(Eq v vnil);
intros;
Refine vnil_lemma zero;
Save vect_zero_eq_vnil;

Goal vect_head : {n|nat}{m:vect (suc n)}A;
intros;
Refine vect_elim'_lemma n m ([_:vect (suc n)]A);
intros;
Immed;
SaveUnfrozen vect_head;

(* Head of a non-empty vector *)

Goal vect_tail : {n|nat}{m:vect (suc n)}(vect n);
intros;
Refine vect_elim'_lemma n m ([_:vect (suc n)](vect n));
intros;
Immed;
SaveUnfrozen vect_tail;

Goal cons_inj_lemma1 : {a,b:A}{n:nat}{v:vect n}(Eq (vcons a v) (vcons b v))
          -> (Eq a b);
intros;
Refine Eq_subst H [x:vect (suc n)](Eq a (vect_head x));
Refine Eq_refl;
Save cons_inj_lemma1;


Goal cons_inj_lemma2 : {a:A}{n:nat}{v,w:vect n}(Eq (vcons a v) (vcons a w))
          -> (Eq v w);
intros;
Refine Eq_subst H [x:vect (suc n)](Eq v (vect_tail x));
Refine Eq_refl;
Save cons_inj_lemma2;


Goal vect_head_cons_tail : {n|nat}{v:vect (suc n)}(Eq (vcons (vect_head v) (vect_tail v)) v);
intros;
Refine vect_elim'_lemma n v [v:vect (suc n)]
           (Eq (vcons (vect_head v) (vect_tail v)) v);
intros;
Refine Eq_refl;
Save vect_head_cons_tail;

Goal vect_nth : {n|nat}{x:fin n}{v:vect n}A;
Refine nat_elim ([n:nat]{x:fin n}{v:vect n}A);
intros;
Refine fin_elim_zero ([_:fin zero]A);
Refine x;
Intros __;
Refine fin_rec';
Refine vect_head;
intros __ ;
Refine x1_ih m;
Refine vect_tail v;
SaveUnfrozen vect_nth;

(*
zeroth == (f_zero two) : (fin three);
first == f_suc (f_zero one) : (fin three);
second == f_suc (f_suc (f_zero zero)): (fin three);
[a,b,c,d:A];
test_vect == vcons a (vcons b (vcons c vnil)) : vect three;
test_vect2 == vcons d (vcons a (vcons c vnil)) : vect three;
vect_nth three first test_vect;
Normal VReg;
vect_nth three zeroth test_vect;
Normal VReg;
vect_nth three second test_vect;
Normal VReg;

vect_nth three first test_vect2;
Normal VReg;
vect_nth three zeroth test_vect2;
Normal VReg;
vect_nth three second test_vect2;
Normal VReg;
*)



Goal vect_zip : {n:nat}{bin:A->A->A}(vect n) -> (vect n) -> (vect n);
intros __;
Refine nat_elim ([n:nat](vect n)->(vect n)->vect n);
intros;
Refine vnil;
intros;
Refine vcons;
Refine bin (vect_head H) (vect_head H1);
Refine x1_ih;
Refine vect_tail H; 
Refine vect_tail H1; 
Save vect_zip;


(*
[a,b,c,d:A];
[bin:A->A->A];

test1 == vcons a (vcons c vnil);
test2 == vcons b (vcons d vnil);
vect_zip two bin test1 test2;
Normal VReg;
*)
(*
Inductive [list:Type]
Parameters [B:Type]
Constructors
[nil:list]
[cons:{a:B}{l:list}list];
Discharge B;
*)

Goal vector_to_list : {n|nat}(vect n) -> (list A);
Refine vect_elim ([n|nat][v:vect n]list A);
Refine nil A;
intros;
Refine cons|A a v_ih;
SaveUnfrozen vector_to_list;

Goal vector_to_list_resp_cons : {n|nat}{v:vect n}{a:A}(Eq (cons|A a (vector_to_list v))
                               (vector_to_list (vcons a v)));
Refine vect_ind [n|nat][v:vect n]({a:A}Eq (cons|A a (vector_to_list v)) (vector_to_list (vcons a v)));
intros;
Refine Eq_refl;
intros;
Refine Eq_resp;
Refine v_ih;
Save vector_to_list_resp_cons;


Goal vect_nth_equiv_nth : {n|nat}{x:fin n}{v:vect n}{a:A}
       (Eq (vect_nth x v) (nth (fin_to_nat x) (vector_to_list v) a));
intros _;
Refine nat_ind [n:nat]{x:fin n}{v:vect n}{a:A}(Eq (vect_nth x v) (nth (fin_to_nat x) (vector_to_list v) a));
intros;
Refine zero_set_is_empty x;
intros;
Qrepl Eq_sym (vect_head_cons_tail v);
Qrepl Eq_sym (vector_to_list_resp_cons (vect_tail v) (vect_head v));
Refine finite_is_zero_or_succ n1 x;
intros;
Qrepl H1;
Refine Eq_refl;
intros;
Refine H1; intros;
Qrepl H2;
Qrepl Eq_sym (fin_to_nat_resp_suc t);
Refine H;
Save vect_nth_equiv_nth;

Discharge A ?;



